/*
 * AUDIO.C
 *
 * (c)Copyright 2002, Matthew Dillon, see COPYRIGHT file.
 * $Backplane: dtv/audio.c,v 1.9 2002/02/11 03:01:23 dillon Exp $
 */

#include "defs.h"

#define DEFAULT_FREQ	8000

Prototype int AudioPassThruInit(Audio *audio);
Prototype int AudioRecordInit(Audio *audio);
Prototype int AudioPlayInit(Audio *audio);
Prototype void AudioFrameGet(Audio *audio);
Prototype void AudioFramePut(Audio *audio, int trigger);
Prototype void AudioFrameDecompress(Audio *audio, int n);
Prototype int AudioFrameCompress(Audio *audio);

static void AudioFree(Audio *audio, int all);

static int
audioInit(Audio *audio)
{
    int v;

    bzero(audio, sizeof(Audio));
    audio->au_CompBufSize = 32768;
    audio->au_CompBuf = malloc(audio->au_CompBufSize);
    audio->au_MixerFd = -1;
    audio->au_DspFd = -1;
    if ((audio->au_MixerFd = open("/dev/mixer", O_RDWR)) < 0) {
	AudioFree(audio, 0);
	return(-1);
    }
    if ((audio->au_DspFd = open("/dev/dsp", O_RDWR)) < 0) {
	AudioFree(audio, 0);
	return(-1);
    }
    v = 0;
    ioctl(audio->au_DspFd, SNDCTL_SEQ_RESET, &v);

    v = AFMT_U8;
    ioctl(audio->au_DspFd, SNDCTL_DSP_SETFMT, &v);
    v = 1;
    ioctl(audio->au_DspFd, SOUND_PCM_WRITE_CHANNELS, &v);
#if 1
    v = 0;
    ioctl(audio->au_DspFd, SNDCTL_DSP_GETBLKSIZE, &v);
    fprintf(stderr, "BLOCKSIZE %d\n", v);
#endif
    audio->au_Good = 1;
    return(0);
}

int
AudioPassThruInit(Audio *audio)
{
    if (audioInit(audio) == 0) {
	return(0);
    }
    return(-1);
}

int
AudioRecordInit(Audio *audio)
{
    if (audioInit(audio) == 0) {
	int v;

#if 1
	fcntl(audio->au_DspFd, F_SETFL, O_NONBLOCK);
#endif

	v = DEFAULT_FREQ;
	ioctl(audio->au_DspFd, SOUND_PCM_WRITE_RATE, &v);
#if 1
	v = 255;
	if (ioctl(audio->au_MixerFd, SOUND_MIXER_WRITE_IGAIN, &v) < 0)
		exit(1);
#endif
	v = SOUND_MASK_LINE;
	ioctl(audio->au_MixerFd, SOUND_MIXER_WRITE_RECSRC, &v);
	return(0);
    }
    return(-1);
}

/*
 * AudioPlayInit() - setup for audio playing
 */

int
AudioPlayInit(Audio *audio)
{
    int v;

    if (audioInit(audio) == 0) {
	struct audio_buf_info info;

	v = DEFAULT_FREQ - 1;
	ioctl(audio->au_DspFd, SOUND_PCM_WRITE_RATE, &v);
	v = 80;
	ioctl(audio->au_MixerFd, SOUND_MIXER_WRITE_VOLUME, &v);
	ioctl(audio->au_DspFd, SNDCTL_DSP_GETOSPACE, &info);
	audio->au_OBytes = info.bytes;
	return(0);
    }
    return(-1);
}

static void
AudioFree(Audio *audio, int all)
{
    if (audio->au_DspFd >= 0) {
	close(audio->au_DspFd);
	audio->au_DspFd = -1;
    }
    if (audio->au_MixerFd >= 0) {
	close(audio->au_MixerFd);
	audio->au_MixerFd = -1;
    }
    if (all) {
	if (audio->au_CompBuf != NULL) {
	    free(audio->au_CompBuf);
	    audio->au_CompBuf = NULL;
	    audio->au_CompBufSize = 0;
	}
    }
}

void
AudioFrameGet(Audio *audio)
{
    int n;

    if (audio->au_Good) {
	n = read(audio->au_DspFd, audio->au_CompBuf, audio->au_CompBufSize);
	if (n < 0)
	    n = 0;
    } else {
	n = 0;
    }
    audio->au_CompIndex = n;
}

void
AudioFramePut(Audio *audio, int trigger)
{
    int xmin = (unsigned char)audio->au_CompBuf[0];
    int xmax = (unsigned char)audio->au_CompBuf[0];
    int j;
    static int Count;

    for (j = 0; j < audio->au_CompIndex; ++j) {
	if (xmin > (unsigned char)audio->au_CompBuf[j])
	    xmin = (unsigned char)audio->au_CompBuf[j];
	if (xmax < (unsigned char)audio->au_CompBuf[j])
	    xmax = (unsigned char)audio->au_CompBuf[j];
	/* audio->au_CompBuf[j] = (j & 2) ? 0xFF : 0x00; */
    }
    if (audio->au_Good) {
	int left;
	int i;
	int skip = audio->au_CompIndex;

	ioctl(audio->au_DspFd, SNDCTL_DSP_GETODELAY, &left);
	/* left = audio->au_OBytes - info.bytes; */
	left &= 16383;	/* XXX broken audio driver */

	/*
	 * If we start building up a backlog, adjust by removing samples.
	 */
	if (trigger != 0 && left > 128)
	    skip = skip / 2;
	if (trigger != 0 && left > 512)
	    skip = skip / 2;
	if (trigger != 0 && left > 1024)
	    skip = skip / 2;
	for (i = 0; i < audio->au_CompIndex; i += skip + 1) {
	    int n;

	    if ((n = audio->au_CompIndex - i) > skip)
		n = skip;
	    if (n)
		write(audio->au_DspFd, audio->au_CompBuf + i, n);
	    Count += n;
	}
	if (trigger > 0 || (trigger < 0 && left > 1024)) {
	    char c = 'S';

	    ioctl(audio->au_DspFd, SNDCTL_DSP_POST, NULL);
	    write(2, &c, 1);
	    Count = 0;
	}
    }
}

void
AudioFrameDecompress(Audio *audio, int n)
{
    audio->au_CompIndex = n;
}

int
AudioFrameCompress(Audio *audio)
{
    return(audio->au_CompIndex);
}

