/*
 * COMP1.C - Compression / Decompression
 *
 * ALGORITHM:
 *
 *	00xxxxxx	6 bit absolute
 *	01xxxyyy	3 bit relative delta for two pixels (-4 to +3)
 *	10rrrrrr	repeat last pixel N times
 *	11rrrrrr	reuse N pixels (no changes)
 *
 * (c)Copyright 2002, Matthew Dillon, see COPYRIGHT file.
 * $Backplane: dtv/comp1.c,v 1.3 2002/02/11 03:01:23 dillon Exp $
 */

#include "defs.h"

Prototype void TVFrameCompress1(TVCap *tv);
Prototype void TVFrameDecompress1(TVCap *tv, XInfo *x, int n);

#define SNAPSIZE	1

static unsigned char *compressrow32(unsigned char *cbuf, unsigned char *cmpbuf, unsigned char *capbuf, int bytes, int skip);
static unsigned char *decompressrow32(unsigned char *cbuf, unsigned char *img, int bytes);

static int XCount;
static int XAvg;

void
TVFrameCompress1(TVCap *tv)
{
    int row;
    int skip;
    int frame = tv->tv_FrameNo % NFRAME;
    unsigned char *cbase = tv->tv_CompBufAry[frame];

    /*sw = 1 - sw;*/

    /*
     * Delta compress R, G, B
     */
    skip = 1 << tv->tv_Shift;
    XCount = 0;
    XAvg = 0;
    for (row = 0; row < tv->tv_YSize; row += skip) {
	int bytes = tv->tv_XSize * tv->tv_PixSize;
	int i = row * tv->tv_XSize * tv->tv_PixSize;
	int ci = row * tv->tv_XSize * MAXPIXSIZE;
	unsigned char *cbuf;
	unsigned char *cmpbuf = tv->tv_CmpBuf;

	switch(tv->tv_PixSize) {
	case 2:
	    assert(0);
	    break;
	case 4:
	    cbuf = compressrow32(cbase + 2, cmpbuf + ci + 0 * tv->tv_XSize, tv->tv_CapBuf + i + 0, bytes, skip);
	    *(u_int16_t *)cbase = cbuf - cbase;
	    cbase = cbuf;
	    cbuf = compressrow32(cbase + 2, cmpbuf + ci + 1 * tv->tv_XSize, tv->tv_CapBuf + i + 1, bytes, skip);
	    *(u_int16_t *)cbase = cbuf - cbase;
	    cbase = cbuf;
	    cbuf = compressrow32(cbase + 2, cmpbuf + ci + 2 * tv->tv_XSize, tv->tv_CapBuf + i + 2, bytes, skip);
	    *(u_int16_t *)cbase = cbuf - cbase;
	    cbase = cbuf;
	    break;
	}
    }
    fprintf(stderr, "%d %d %d avg=%d\n", tv->tv_FrameNo, cbase - tv->tv_CompBufAry[frame], XCount, XCount / (XAvg + 1));
    tv->tv_CompCountAry[frame] = cbase - tv->tv_CompBufAry[frame];
}

static __inline u_int8_t
cvalue32(unsigned char *capbuf, int skip)
{
    u_int8_t cvalue;

    switch(skip) {
    case 1:
	cvalue = capbuf[0] >> 2;
	break;
    case 2:
	cvalue = (capbuf[0] + capbuf[4] + 1) >> 3;
	break;
    case 3:
	cvalue = (capbuf[0] + capbuf[4] + capbuf[8] + 2) >> 4;
	break;
    case 4:
    default:
	cvalue = (capbuf[0] + capbuf[4] + capbuf[8] + capbuf[12] + 3) >> 5;
	break;
    }
    return(cvalue);
}

static unsigned char *
compressrow32(unsigned char *cbuf, unsigned char *cmpbuf, unsigned char *capbuf, int bytes, int skip)
{
    int xskip = 4 * skip;
    unsigned char *ebuf = capbuf + bytes - xskip;
    u_int8_t lvalue1;
    u_int8_t lvalue2;

    /*
     * note: cmpbuf is synchronized with cbuf, whereas capbuf is one pixel
     * ahead.  data in cmpbuf is stored pre-shifted (>> 2).
     */
    lvalue2 = cvalue32(capbuf, skip);	/* prime lvalue2 */
    *cbuf++ = lvalue2;
    cmpbuf[0] = lvalue2;
    ++cmpbuf;
    capbuf += xskip;

    lvalue1 = cvalue32(capbuf, skip);	/* prime lvalue1 */
    capbuf += xskip;

    while (capbuf < ebuf) {
	u_int8_t cvalue;
	int8_t delta1;
	int8_t delta2;

	cvalue = cvalue32(capbuf, skip);	/* rip next pixel value */
	capbuf += xskip;

#if 1
	/*
	 * Check pixel deltas against previous frame
	 */
	delta2 = lvalue1 - cmpbuf[0];		/* delta against prior frame */
	delta1 = cvalue - cmpbuf[1];	/* delta against prior frame */
	if (
	    delta1 >= -SNAPSIZE && delta2 >= -SNAPSIZE && 
	    delta1 < SNAPSIZE && delta2 < SNAPSIZE
	) {
	    int count = 2;

	    ++cmpbuf;			/* eat lvalue1 */
	    ++cmpbuf;			/* eat cvalue */
	    /* cmpbuf now synchronized with capbuf */

	    while (capbuf < ebuf && count < 0x3F) {
		cvalue = cvalue32(capbuf, skip);
		delta1 = cvalue - cmpbuf[0];
		if (delta1 < -SNAPSIZE || delta1 > SNAPSIZE)
		    break;
		++cmpbuf;		/* keep prior cmpbuf value */
		capbuf += xskip;	/* eat cvalue */
		++count;
	    }
	    *cbuf++ = 0xC0 | count;
	    lvalue2 = cmpbuf[-1];	/* last completed value */
	    lvalue1 = cvalue32(capbuf, skip);	/* reprime lvalue1 */
	    capbuf += xskip;	/* eat cvalue */
	    continue;
	}
#endif

	/*
	 * Check sequential pixel deltas
	 */
	delta2 = lvalue1 - lvalue2;		/* delta against prior pixel */
	delta1 = cvalue - lvalue1;		/* delta against prior pixel */

#if 1
	if (
	    delta2 >= -SNAPSIZE && delta2 < SNAPSIZE && 
	    delta1 + delta2 >= -SNAPSIZE && delta1 + delta2 < SNAPSIZE
	) {
	    int count = 2;

	    cmpbuf[0] = lvalue2;	/* eat lvalue1 */
	    ++cmpbuf;
	    cmpbuf[0] = lvalue2;	/* eat cvalue */
	    ++cmpbuf;

	    while (capbuf < ebuf && count < 0x3F) {
		cvalue = cvalue32(capbuf, skip);
		delta1 = cvalue - lvalue2;
		if (delta1 < -SNAPSIZE || delta1 > SNAPSIZE)
		    break;
		cmpbuf[0] = lvalue2;
		++cmpbuf;
		capbuf += xskip;	/* eat cvalue */
		++count;
	    }
	    *cbuf++ = 0x80 | count;
	    cvalue = cvalue32(capbuf, skip);
	    capbuf += xskip;	/* eat cvalue */
	    lvalue1 = cvalue;
	    continue;
	}
#endif

	/*
	 * Store as delta.
	 */
	if (delta1 >= -4 && delta2 >= -4 && delta1 < 4 && delta2 < 4) {
	    *cbuf++ = 0x40 | ((delta2 & 7) << 3) | (delta1 & 7);
	    cmpbuf[0] = lvalue1;
	    ++cmpbuf;
	    cmpbuf[0] = cvalue;
	    ++cmpbuf;

	    lvalue2 = cvalue;
	    lvalue1 = cvalue32(capbuf, skip);
	    capbuf += xskip;
	    continue;
	}

	/*
	 * Unable to optimize, store lvalue1 as 6 bit color value
	 */
	*cbuf++ = lvalue1;
	cmpbuf[0] = lvalue1;
	++cmpbuf;
	lvalue2 = lvalue1;
	lvalue1 = cvalue;
    }
    *cbuf++ = lvalue1;		/* eat lvalue1 */
    cmpbuf[0] = lvalue1;
    ++cmpbuf;

    while (capbuf < ebuf + xskip) {
	u_int8_t cvalue = cvalue32(capbuf, skip);
	capbuf += xskip;
	*cbuf++ = cvalue;
	cmpbuf[0] = cvalue;
	++cmpbuf;
    }
    while ((int)cbuf & 1)
	++cbuf;
    return(cbuf);
}

/*
 * Decompress the input buffer.  Do not bother
 * decompressing rows that we are not going to
 * use.
 */
void
TVFrameDecompress1(TVCap *tv, XInfo *x, int n)
{
    int row;
    int mult = 1 << tv->tv_Shift;
    int frame = tv->tv_FrameNo % NFRAME;
    unsigned char *cbase = tv->tv_CompBufAry[frame];

    for (row = 0; row < tv->tv_YSize; ++row) {
	int bytes = tv->tv_XSize * tv->tv_PixSize;
	unsigned char *cbuf;
	unsigned char *obuf;

	if (x->Flags & XF_DIRECT)
	    obuf = x->x_Image->data + row * x->x_Image->bytes_per_line;
	else
	    obuf = tv->tv_CmpBuf + row * bytes;

	if (row % mult == 0) {
	    if (tv->tv_PixSize == 2) {
		assert(0);
	    } else {
		cbuf = decompressrow32(cbase + 2, obuf + 0, bytes);
		assert(cbase + *(u_int16_t *)cbase == cbuf);
		cbase = cbuf;
		cbuf = decompressrow32(cbase + 2, obuf + 1, bytes);
		assert(cbase + *(u_int16_t *)cbase == cbuf);
		cbase = cbuf;
		cbuf = decompressrow32(cbase + 2, obuf + 2, bytes);
		assert(cbase + *(u_int16_t *)cbase == cbuf);
		cbase = cbuf;
	    }
	} else {
	    cbase += *(u_int16_t *)cbase;
	    if (tv->tv_PixSize == 4) {
		cbase += *(u_int16_t *)cbase;
		cbase += *(u_int16_t *)cbase;
	    }
	}
    }
}

static
unsigned char *
decompressrow32(unsigned char *cbuf, unsigned char *img, int bytes)
{
    unsigned char *eimg = img + bytes;
    u_int8_t lvalue = 0;

    while (img < eimg) {
	u_int8_t cvalue = *cbuf++;

	switch(cvalue & 0xC0) {
	case 0x00:
	    lvalue = cvalue;
	    *img = cvalue << 2;
	    img += 4;
	    break;
	case 0x40:
	    lvalue = lvalue + ((int8_t)(cvalue << 2) >> 5);
	    *img = lvalue << 2;
	    img += 4;
	    lvalue = lvalue + ((int8_t)(cvalue << 5) >> 5);
	    *img = lvalue << 2;
	    img += 4;
	    break;
	case 0x80:
	    cvalue &= 0x3F;
	    while (cvalue) {
		*img = lvalue << 2;
		img += 4;
		--cvalue;
	    }
	    break;
	case 0xC0:
	    img += 4 * (cvalue & 0x3F);
	    lvalue = img[-4] >> 2;
	    break;
	}
    }
    while ((int)cbuf & 1)
	++cbuf;
    return (cbuf);
}

