/*
 * COMP2.C - Compression / Decompression
 *
 * ALGORITHM:
 *	00rrggbb		delta prior pixel +2,+1,-1,-2
 *	01rrggbb		delta underneath pixel +2,+1,-1,-2
 *	10rrrrgg ggggbbbb	absolute
 *	11cccccc		skip C pixels
 *
 * (c)Copyright 2002, Matthew Dillon, see COPYRIGHT file.
 * $Backplane: dtv/comp2.c,v 1.4 2002/02/11 03:01:23 dillon Exp $
 */

#include "defs.h"

Prototype void TVFrameCompress2(TVCap *tv);
Prototype void TVFrameDecompress2(TVCap *tv, XInfo *x, int n);

#ifndef DSHIFT
#define DSHIFT		1	/* preshift (1 or 2) */
#define SS		4	/* snap size */
#endif

#ifndef TVFRAMECOMPRESS
#define TVFRAMECOMPRESS		TVFrameCompress2
#define TVFRAMEDECOMPRESS	TVFrameDecompress2
#endif

static unsigned char *compressrow32(unsigned char *cbuf, unsigned char *cmpbuf, unsigned char *capbuf, int bytes, int skip);
static unsigned char *decompressrow32(unsigned char *cbuf, unsigned char *img, int bytes);
static void InitUndo(void);

static int XCount;
static int XAvg;

static unsigned char UndoR[256];
static unsigned char UndoG[256];
static unsigned char UndoB[256];
static int UndoSetup;

void
TVFRAMECOMPRESS(TVCap *tv)
{
    int row;
    int skip;
    int frame = tv->tv_FrameNo % NFRAME;
    unsigned char *cbase = tv->tv_CompBufAry[frame];

    if (UndoSetup == 0)
	InitUndo();

    /*
     * Delta compress R, G, B
     */
    skip = 1 << tv->tv_Shift;
    XCount = 0;
    XAvg = 0;
    for (row = 0; row < tv->tv_YSize; row += skip) {
	int bytes = tv->tv_XSize * tv->tv_PixSize;
	int i = row * tv->tv_XSize * tv->tv_PixSize;
	int ci = row * tv->tv_XSize * MAXPIXSIZE;
	unsigned char *cbuf;
	unsigned char *cmpbuf = tv->tv_CmpBuf;

	switch(tv->tv_PixSize) {
	case 2:
	    assert(0);
	    break;
	case 4:
	    cbuf = compressrow32(cbase + 2, cmpbuf + ci, tv->tv_CapBuf + i, bytes, skip);
	    *(u_int16_t *)cbase = cbuf - cbase;
	    cbase = cbuf;
	    break;
	}
    }
    fprintf(stderr, "%d %d %d avg=%d\n", tv->tv_FrameNo, cbase - tv->tv_CompBufAry[frame], XCount, XCount / (XAvg + 1));
    tv->tv_CompCountAry[frame] = cbase - tv->tv_CompBufAry[frame];
}

#define ENCODE(var, delta_eq)		\
    {					\
	int8_t delta = delta_eq;	\
	if (delta < -2) {		\
	    var = -2;			\
	} else if (delta >= 2) {	\
	    var = 0;			\
	} else if (delta == 0) {	\
	    if (var == 0)		\
		var = 1;		\
	    else if (var == (0xFF >> DSHIFT))	\
		var = -1;		\
	    else			\
		var = -1;		\
	} else {			\
	    var = delta;		\
	}				\
    }					\

static unsigned char *
compressrow32(unsigned char *cbuf, unsigned char *cmpbuf, unsigned char *capbuf, int bytes, int skip)
{
    unsigned char *ebuf = capbuf + bytes;
    int xskip = 4 * skip;
    int count = -1;
    unsigned char lr = 0;
    unsigned char lg = 0;
    unsigned char lb = 0;

    while (capbuf < ebuf) {
	unsigned char r;
	unsigned char g;
	unsigned char b;

	switch(skip) {
	case 1:
	    r = capbuf[0];
	    g = capbuf[1];
	    b = capbuf[2];
	    break;
	case 2:
	    r = (capbuf[0] + capbuf[4] + 1) >> 1;
	    g = (capbuf[1] + capbuf[5] + 1) >> 1;
	    b = (capbuf[2] + capbuf[6] + 1) >> 1;
	    break;
	case 3:
	    r = (capbuf[0] + capbuf[4] + capbuf[8] + 1) / 3;
	    g = (capbuf[1] + capbuf[5] + capbuf[9] + 1) / 3;
	    b = (capbuf[2] + capbuf[6] + capbuf[10] + 1) / 3;
	    break;
	case 4:
	default:
	    r = (capbuf[0] + capbuf[4] + capbuf[8] + capbuf[12] + 2) >> 2;
	    g = (capbuf[1] + capbuf[5] + capbuf[9] + capbuf[13] + 2) >> 2;
	    b = (capbuf[2] + capbuf[6] + capbuf[10] + capbuf[14] + 2) >> 2;
	    break;
	}
	r >>= DSHIFT;
	g >>= DSHIFT;
	b >>= DSHIFT;
#if 1
	if ((int8_t)(r - cmpbuf[0]) >= -SS && (int8_t)(r - cmpbuf[0]) <= SS &&
	    (int8_t)(g - cmpbuf[1]) >= -SS && (int8_t)(g - cmpbuf[1]) <= SS &&
	    (int8_t)(b - cmpbuf[2]) >= -SS && (int8_t)(b - cmpbuf[2]) <= SS &&
	    count < 0x3F
	) {
	    unsigned char v;

	    /*
	     * Encode a delta.  If count is > 0 then just fake it so we
	     * can easily back-up later.
	     */
	    if (count <= 0) {
		ENCODE(r, (int8_t)(r - cmpbuf[0]));
		ENCODE(g, (int8_t)(g - cmpbuf[1]));
		ENCODE(b, (int8_t)(b - cmpbuf[2]));
		v = ((r & 3) << 4) |
		    ((g & 3) << 2) |
		    ((b & 3));
		cmpbuf[0] += UndoR[v];
		lr = cmpbuf[0];
		cmpbuf[1] += UndoG[v];
		lg = cmpbuf[1];
		cmpbuf[2] += UndoB[v];
		lb = cmpbuf[2];
		cbuf[0] = v | 0x40;
	    } else {
		cbuf[0] = 0xFF;		/* is ignored */
		lr = cmpbuf[0];
		lg = cmpbuf[1];
		lb = cmpbuf[2];
	    }
	    ++cbuf;
	    ++count;
	} else if (count > 0) {
	    if (count > 1) {
		unsigned char v;

		cbuf -= count;
		v = cbuf[0];
		cmpbuf[-count*4+0] -= UndoR[v];
		cmpbuf[-count*4+1] -= UndoG[v];
		cmpbuf[-count*4+2] -= UndoB[v];
		
		while (count > 0x3F) {
		    *cbuf++ = 0xC0 | 0x3F;
		    count -= 0x3F;
		}
		*cbuf++ = 0xC0 | count;
	    }
	    lr = cmpbuf[-4];
	    lg = cmpbuf[-3];
	    lb = cmpbuf[-2];
	    count = -1;
	    continue;
	} else 
#endif
#if 1
	if ((int8_t)(r - lr) >= -SS && (int8_t)(r - lr) <= SS &&
	    (int8_t)(g - lg) >= -SS && (int8_t)(g - lg) <= SS &&
	    (int8_t)(b - lb) >= -SS && (int8_t)(b - lb) <= SS
	) {
	    unsigned char v;

	    ENCODE(r, (int8_t)(r - lr));
	    ENCODE(g, (int8_t)(g - lg));
	    ENCODE(b, (int8_t)(b - lb));
	    v = ((r & 3) << 4) |
		((g & 3) << 2) |
		(b & 3);
	    lr += UndoR[v];
	    cmpbuf[0] = lr;
	    lg += UndoG[v];
	    cmpbuf[1] = lg;
	    lb += UndoB[v];
	    cmpbuf[2] = lb;
	    cbuf[0] = v;
	    ++cbuf;
	} else 
#endif
	/*
	 * Fallback, jam the color.  The high two bits of cbuf[0] are
	 * reserved for the control code, so we may have to shift 'r' more.
	 * We have to make sure our 'lr' and cmpbuf[0] matches what the
	 * decoder would decode.
	 */
	{
#if DSHIFT < 2
	    cmpbuf[0] = lr = r & ((0xFF >> (2 - DSHIFT)) << (2 - DSHIFT));
#else
	    cmpbuf[0] = lr = r;
#endif
	    cbuf[1] = cmpbuf[1] = lg = g;
	    cbuf[2] = cmpbuf[2] = lb = b;
#if DSHIFT < 2
	    cbuf[0] = 0x80 | (r >> (2 - DSHIFT));
#else
	    cbuf[0] = 0x80 | r;
#endif
	    cbuf += 3;
	}
	capbuf += xskip;
	cmpbuf += 4;
    }
    if (count > 1) {
#if 1
	unsigned char v;
	cbuf -= count;
	v = cbuf[0];
	cmpbuf[-count*4+0] -= UndoR[v];
	cmpbuf[-count*4+1] -= UndoG[v];
	cmpbuf[-count*4+2] -= UndoB[v];
	while (count > 0x3F) {
	    *cbuf++ = 0xC0 | 0x3F;
	    count -= 0x3F;
	}
	*cbuf++ = 0xC0 | count;
#endif
    }
    while ((int)cbuf & 1)
	++cbuf;
    return(cbuf);
}

/*
 * Decompress the input buffer.  Do not bother
 * decompressing rows that we are not going to
 * use.
 */
void
TVFRAMEDECOMPRESS(TVCap *tv, XInfo *x, int n)
{
    int row;
    int mult = 1 << tv->tv_Shift;
    int frame = tv->tv_FrameNo % NFRAME;
    unsigned char *cbase = tv->tv_CompBufAry[frame];

    if (UndoSetup == 0)
	InitUndo();
    for (row = 0; row < tv->tv_YSize; ++row) {
	int bytes = tv->tv_XSize * tv->tv_PixSize;
	unsigned char *cbuf;
	unsigned char *obuf;

	if (x->Flags & XF_DIRECT) {
	    obuf = x->x_Image->data + row * x->x_Image->bytes_per_line;
	} else {
	    obuf = tv->tv_CmpBuf + row * bytes;
	}

	if (row % mult == 0) {
	    if (tv->tv_PixSize == 2) {
		assert(0);
	    } else {
		cbuf = decompressrow32(cbase + 2, obuf + 0, bytes);
		assert(cbase + *(u_int16_t *)cbase == cbuf);
		cbase = cbuf;
	    }
	} else {
	    cbase += *(u_int16_t *)cbase;
	}
    }
}

static
unsigned char *
decompressrow32(unsigned char *cbuf, unsigned char *img, int bytes)
{
    unsigned char *eimg = img + bytes;
    unsigned char lr = 0;
    unsigned char lg = 0;
    unsigned char lb = 0;

    while (img < eimg) {
	unsigned char v = cbuf[0];
	switch(v & 0xC0) {
	case 0x00:
	    lr += UndoR[v];
	    lg += UndoG[v];
	    lb += UndoB[v];
	    ++cbuf;
	    break;
	case 0x40:
	    lr = (img[0] >> DSHIFT) + UndoR[v];
	    lg = (img[1] >> DSHIFT) + UndoG[v];
	    lb = (img[2] >> DSHIFT) + UndoB[v];
	    ++cbuf;
	    break;
	case 0x80:
#if DSHIFT < 2
	    lr = (cbuf[0] & 0x3F) << (2 - DSHIFT);
#else
	    lr = (cbuf[0] & 0x3F);
#endif
	    lg = cbuf[1];
	    lb = cbuf[2];
	    cbuf += 3;
	    break;
	case 0xC0:
	    img += (cbuf[0] & 0x3F) * 4;
	    lr = img[-4] >> DSHIFT;
	    lg = img[-3] >> DSHIFT;
	    lb = img[-2] >> DSHIFT;
	    ++cbuf;
	    continue;
	}
	img[0] = lr << DSHIFT;
	img[1] = lg << DSHIFT;
	img[2] = lb << DSHIFT;
	img += 4;
    }
    while ((int)cbuf & 1)
	++cbuf;
    return(cbuf);
}

static
void
InitUndo(void)
{
    int i;

    for (i = 0; i < 256; ++i) {
	switch((i >> 4) & 3) {
	case 0:		/* +2 */
	    UndoR[i] = 2;
	    break;
	case 1:		/* +1 */
	    UndoR[i] = 1;
	    break;
	case 2:		/* -2 */
	    UndoR[i] = -2;
	    break;
	case 3:		/* -1 */
	    UndoR[i] = -1;
	    break;
	}
	switch((i >> 2) & 3) {
	case 0:		/* +2 */
	    UndoG[i] = 2;
	    break;
	case 1:		/* +1 */
	    UndoG[i] = 1;
	    break;
	case 2:		/* -2 */
	    UndoG[i] = -2;
	    break;
	case 3:		/* -1 */
	    UndoG[i] = -1;
	    break;
	}
	switch(i & 3) {
	case 0:		/* +2 */
	    UndoB[i] = 2;
	    break;
	case 1:		/* +1 */
	    UndoB[i] = 1;
	    break;
	case 2:		/* -2 */
	    UndoB[i] = -2;
	    break;
	case 3:		/* -1 */
	    UndoB[i] = -1;
	    break;
	}
    }
}

