/*
 * MAIN.C
 *
 * (c)Copyright 2002, Matthew Dillon, see COPYRIGHT file.
 * $Backplane: dtv/main.c,v 1.24 2002/02/11 03:13:57 dillon Exp $
 */

#include "defs.h"

#include <sys/socket.h>         /* internet sockets     */
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

Prototype int VerboseOpt;

static int xread(int fd, void *buf, int bytes);
static int xwrite(int fd, void *buf, int bytes);
static int ConnectINetSocket(const char *host, int port);
static void setsockbuffer(int fd, int how);
static void frameDelayInit(void);
static void frameDelay(TVCap *tv);
static void setAlgorithm(TVHead *head, const char **chset);
static void blockFixup(TVBlock *b);
static void blockEncode(TVBlock *b);

#if 0
static void forkinputpipe(int fd);
static void forkoutputpipe(int fd);
static void forkoutputbuffer(int fd);
#endif

void (*TVFrameCompress)(TVCap *tv);
void (*TVFrameDecompress)(TVCap *tv, XInfo *x, int n);

int VerboseOpt;

int
main(int ac, char **av)
{
    TVCap tv;
    XInfo xinfo;
    Audio audio;
    TVHead head;
    char *emsg = NULL;
    char *host = NULL;
    int i;
    int fd;
    int mode = 0;
    int count = 0;
    int shift1 = 0;	/* reduction shift */
    int shift2 = 0;	/* expansion shift */
    int record_mode = 0;
    int trigger;
    int channel = 12;
    const char *chset = NULL;
    int xflags = 0;
    fd_set rfds;

    FD_ZERO(&rfds);

    bzero(&head, sizeof(head));

    head.h_Version = TVH_VERSION;
    head.h_Size = sizeof(head);
    head.h_CAlgorithm = 1;

    if (ac == 1) {
	fprintf(
	    stderr, 
	    "%s " VERSIONSTR ", Copyright 2002, Matthew Dillon\n",
	    av[0]
	);
	fprintf(stderr,
	    "%s [-A#][-i][-s][-c][-w][-h host][-z][-Z][-C channel][-S][-v[#]]\n"
	    "   [-T channelset]\n"
	    "    -A#        Set compression algorithm 1-4 (note1).\n"
	    "    -i         Close stderr, implies server mode.\n"
	    "    -s         Server mode.\n"
	    "    -c         Client mode.\n"
	    "    -w         Recording mode.  Used to have either the client\n"
	    "               or server write out a file.  NOTE: when serving\n"
	    "               from inetd.conf, do not use -w or the client will\n"
	    "               attempt to provide its own frame delays.\n"
	    "    -h host    Connect to dtv server on host.  This implies\n"
	    "               client mode.  If not specified, a client will\n"
	    "               read from stdin.\n"
	    "    -z[z...]   Tell server to pre-reduce the frame size (note 1).\n"
	    "    -Z[Z...]   Tell client to blow-up the frame size (note 1).\n"
	    "    -C chan    Specify channel (note 1).\n"
	    "    -S         When blowing-up frames, do not do line fill.\n"
	    "    -v[#]      Turn on verbose mode w/ an optional level\n"
	    "    -v[v...]   Another way to turn on verbose mode.\n"
	    "    -T chset   Specify source, typically nabcst or cableirc (note 1):\n"
	    "                 cableirc (default), cablehrc, nabcst, weurope,\n"
	    "                 jpnbcst, jpncable, xussr, australia, france\n"
	    "\n"
	    "  (note 1): The client will pass these parameters to the server,\n"
	    "            and may also be specified on the server to change\n"
	    "            the default.\n"
	    "\n"
	    "Verbose Indicators:\n"
	    "    f          Frame received\n"
	    "    T          Lost frame received (server lost frame due to\n"
	    "               backed-up tcp connection or lack of cpu.  Note\n"
	    "               that a number of T's always occur at the\n"
	    "               beginning for sound-buffering purposes.\n"
	    "    S          Lost sound synchronization.  Synchronization\n"
	    "               should slowly recover.\n"
	    "\n"
	    "inetd.conf: mondex stream tcp nowait someuser <path>/dtv dtv -s -i\n"
	    "server-record:   dtv -C 12 -s -w -A2 -z > file\n"
	    "client-record:   dtv -C 12 -h server -w -A2 -z > file\n"
	    "client-realtime: dtv -C 12 -h server -A2 -z -Z\n"
	    "standalone TV:   dtv -C 12 -A2 -z\n"
	    "\n"
	    "   -A3 and -A4 will give higher compressions.  One typically also\n"
	    "   uses -z (client or server) to cut the frame size, increase\n"
	    "   compression yields, and reduce cpu.  -Z is often used to make\n"
	    "   the window larger on the client (if it can handle it).  When\n"
	    "   networking, 100BaseT full-duplex is recommended.  A fast\n"
	    "   graphics card like a GeForce2 MX-200 is recommended, and a\n"
	    "   700MHz or faster cpu.  -S may reduce cpu consumption on the\n"
	    "   client when used in conjuction with -Z.\n"
	    , av[0]
	);
	exit(1);
    }

    for (i = 1; i < ac; ++i) {
	char *ptr = av[i];
	if (*ptr != '-') {
	    fprintf(stderr, "Unexpected argument: %s\n", ptr);
	    exit(1);
	}
	ptr += 2;
	switch(ptr[-1]) {
	case 'T':
	    ptr = (*ptr) ? ptr : av[++i];
	    chset = ptr;
	    break;
	case 'A':
	    ptr = (*ptr) ? ptr : av[++i];
	    head.h_CAlgorithm = strtol(ptr, NULL, 0);
	    break;
	case 'i':
	    freopen("/dev/null", "w", stderr);
	    /* fall through */
	case 's':
	    mode = 1;
	    break;
	case 'w':
	    record_mode = 1;
	    break;
	case 'c':
	    mode = 2;
	    break;
	case 'h':
	    mode = 2;
	    host = (*ptr) ? ptr : av[++i];
	    break;
	case 'Z':	/* make the client-side frame larger */
	    --shift2;
	    while (*ptr == 'Z') {
		--shift2;
		++ptr;
	    }
	    break;
	case 'z':	/* make the server-side frame smaller */
	    ++shift1;
	    while (*ptr == 'z') {
		++shift1;
		++ptr;
	    }
	    break;
	case 'C':
	    ptr = (*ptr) ? ptr : av[++i];
	    channel = strtol(ptr, NULL, 0);
	    break;
	case 'S':
	    xflags |= XF_SKIPLINES;
	    break;
	case 'v':
	    ++VerboseOpt;
	    while (*ptr == 'v') {
		++VerboseOpt;
		++ptr;
	    }
	    if (*ptr >= '0' && *ptr <= '9')
		VerboseOpt = strtol(ptr, NULL, 0);
	    break;
	default:
	    fprintf(stderr, "Unknown option: %s\n", ptr - 2);
	    exit(1);
	}
    }
    switch(mode) {
    case 0:	/* server and client */
	setAlgorithm(&head, &chset);
	TVCapInit(&tv, shift1 + shift2, channel, chset);
	if (TVCapStart(&tv, &emsg)) {
	    fprintf(stderr, "%s\n", emsg);
	    exit(1);
	}
	FrameInit(&tv, &xinfo, shift1 + shift2, xflags|XF_NODIRECT);
	FrameOpen(&tv, &xinfo);
	FrameUserControl(&tv, &xinfo, 1);
	AudioPassThruInit(&audio);
	for (;;) {
	    if (TVFrameGet(&tv) != 0) {
		FrameConvert(&tv, &xinfo);
		FrameDisplay(&xinfo);
		if (VerboseOpt)
		    write(2, "f", 1);
	    } else {
		if (VerboseOpt)
		    write(2, "T", 1);
	    }
	    ++count;
	}
	break;
    case 1:	/* server only */
	/*
	 * If we are not in record mode we are interacting with a client.
	 */
	if (record_mode == 0)
	    xread(0, &head, sizeof(head));
	else
	    head.h_Flags |= THF_FILE;

	head.h_Shift += shift1 + shift2;
	if (head.h_Shift < 0)	/* silly to expand from 1:1 on server */
	    head.h_Shift = 0;

	setAlgorithm(&head, &chset);

	TVCapInit(&tv, head.h_Shift, head.h_Channel, chset);
	if (TVCapStart(&tv, &emsg)) {
	    fprintf(stderr, "%s\n", emsg);
	    exit(1);
	}
	setsockbuffer(1, SO_SNDBUF);
#if 0
	forkoutputpipe(1);
	forkoutputbuffer(1);
	forkoutputbuffer(1);
#endif
	AudioRecordInit(&audio);

	head.h_XSize = tv.tv_XSize >> tv.tv_Shift;
	head.h_YSize = tv.tv_YSize >> tv.tv_Shift;

	xwrite(1, &head, sizeof(head));

	TVFrameGet(&tv);	/* prime the frame retrieval */
	AudioFrameGet(&audio);	/* prime the audio */

	for (;;) {
	    int r;
	    int tvn;
	    TVBlock block;
	    struct timeval timo = { 0, 0 };

	    /*
	     * Get the TV and audio frame.  If we missed the TV frame
	     * tvn will be 0.
	     */
	    tvn = TVFrameGet(&tv);
	    AudioFrameGet(&audio);

	    /*
	     * Handle controls
	     */
	    FD_SET(0, &rfds);
	    select(1, &rfds, NULL, NULL, &timo);
	    if (FD_ISSET(0, &rfds)) {
		char buf[32];
		if (fgets(buf, sizeof(buf), stdin) != NULL) {
		    if (strncmp(buf, "CHUP", 4) == 0) {
			++tv.tv_Channel;
		    } else if (strncmp(buf, "CHDN", 4) == 0) {
			if (tv.tv_Channel > 1)
			    --tv.tv_Channel;
		    } else if (strncmp(buf, "CHST", 4) == 0) {
			tv.tv_Channel = strtol(buf + 4, NULL, 10);
			if (tv.tv_Channel < 2)
			    tv.tv_Channel = 2;
			if (tv.tv_Channel > 1000)
			    tv.tv_Channel = 1000;
		    }
		    TVCapSetChannel(&tv);
		    block.b_Type = BT_CCHANGE;
		    block.b_Info = tv.tv_Channel;
		    block.b_Size = 0;
		    blockEncode(&block);
		    if (xwrite(1, &block, sizeof(block)) != sizeof(block))
			break;
		}
	    }

	    /*
	     * Process the audio frame
	     */
	    block.b_Type = BT_AUDIO;
	    block.b_Info = 0;
	    block.b_Size = AudioFrameCompress(&audio);
	    blockEncode(&block);
	    if (xwrite(1, &block, sizeof(block)) != sizeof(block))
		break;
	    blockFixup(&block);
	    if ((r = xwrite(1, audio.au_CompBuf, block.b_Size)) != block.b_Size) {
		fprintf(stderr, "only wrote %d/%d\n", r, block.b_Size);
		break;
	    }

	    /*
	     * Process the TV frame.  Do not bump the frame counter on
	     * the server side, we do not have to delay frames.
	     */
	    if (tvn) {
		TVFrameCompress(&tv);
		block.b_Type = BT_FRAME;
		block.b_Info = 0;
		block.b_Size = tv.tv_CompCountAry[0];
		blockEncode(&block);
		if (xwrite(1, &block, sizeof(block)) != sizeof(block))
		    break;
		blockFixup(&block);
		if ((r = xwrite(1, tv.tv_CompBufAry[0], block.b_Size)) != block.b_Size) {
		    fprintf(stderr, "only wrote %d/%d\n", r, block.b_Size);
		    break;
		}
	    } else {
		block.b_Type = BT_FRAME;
		block.b_Info = 0;
		block.b_Size = 0;
		blockEncode(&block);
		if (xwrite(1, &block, sizeof(block)) != sizeof(block))
		    break;
	    }
	    ++count;
	}
	break;
    case 2:	/* client only */
	/*
	 * Setup initialization request to send to the server.  Pass
	 * reduction request to server, leave expansion request to the
	 * client.
	 */
	head.h_Shift = shift1;
	head.h_Channel = channel;
	setAlgorithm(&head, &chset);

	if (host) {
	    if ((fd = ConnectINetSocket(host, 471)) < 0) {
		fprintf(stderr, "Cannot connect to %s\n", host);
		exit(1);
	    }
	    xwrite(fd, &head, sizeof(head));
	} else {
	    /*
	     * batching from file, client may want to reduce the scale
	     */
	    fd = 0;
	    shift2 += shift1;
	}
	setsockbuffer(fd, SO_RCVBUF);
#if 0
	forkinputpipe(fd);
#endif

	xread(fd, &head, sizeof(head));
	setAlgorithm(&head, &chset);
	if (record_mode != 0) {
	    int oflags = head.h_Flags;
	    head.h_Flags |= THF_FILE;
	    if (xwrite(1, &head, sizeof(head)) != sizeof(head))
		break;
	    head.h_Flags = oflags;
	}
	if (VerboseOpt)
	    fprintf(stderr, "Frame %dx%d\n", head.h_XSize, head.h_YSize);

	/*
	 * XXX get channel set from server
	 */
	TVCapInitSlave(&tv, head.h_XSize, head.h_YSize, shift2, -1, chset);
	AudioPlayInit(&audio);
	if (record_mode == 0) {
	    FrameInit(&tv, &xinfo, shift2, xflags);
	    FrameOpen(&tv, &xinfo);
	    tv.tv_Channel = head.h_Channel;	/* XXX for FrameUserControl */
	    FrameUserControl(&tv, &xinfo, fd);
	}

	trigger = 0;

	if (record_mode == 0) {
	    if (head.h_Flags & THF_FILE)
		frameDelayInit();
	}
	for (;;) {
	    /*
	     * Audio Frame
	     */
	    char c;
	    int frame;
	    int n;
	    TVBlock block;

	    if (xread(fd, &block, sizeof(block)) != sizeof(block))
		break;
	    if (record_mode != 0) {
		if (xwrite(1, &block, sizeof(block)) != sizeof(block))
		    break;
	    }
	    blockFixup(&block);
	    if (block.b_Type != BT_AUDIO) {
		if (block.b_Type == BT_CCHANGE)
		    ;
		continue;
	    }
	    if (block.b_Size < 0 || block.b_Size > audio.au_CompBufSize) {
		fprintf(
		    stderr, 
		    "bad audio frame size: %d/%d\n", 
		    block.b_Size, 
		    audio.au_CompBufSize
		);
		break;
	    }
	    if (xread(fd, audio.au_CompBuf, block.b_Size) != block.b_Size)
		break;
	    if (trigger == 0 && tv.tv_FrameNo == NFRAME)
		trigger = 1;
	    if (record_mode == 0) {
		AudioFrameDecompress(&audio, block.b_Size);
		AudioFramePut(&audio, trigger);
	    } else {
		if (xwrite(1, audio.au_CompBuf, block.b_Size) != block.b_Size)
		    break;
	    }
	    if (trigger > 0)
		trigger = -1;

	    /*
	     * Video Frame
	     */
	    frame = tv.tv_FrameNo % NFRAME;
	    if (xread(fd, &block, sizeof(block)) != sizeof(block))
		break;
	    if (record_mode != 0) {
		if (xwrite(1, &block, sizeof(block)) != sizeof(block))
		    break;
	    }
	    blockFixup(&block);
	    if (VerboseOpt > 1)
		fprintf(stderr, " %6d\n", block.b_Size);
	    if (block.b_Size < 0 || block.b_Size > tv.tv_CompBufSize) {
		fprintf(stderr, "bad video frame size: %d\n", block.b_Size);
		break;
	    }
	    tv.tv_CompCountAry[frame] = block.b_Size;
	    if (block.b_Size) {
		if (xread(fd, tv.tv_CompBufAry[frame], block.b_Size) != block.b_Size)
		    break;
		if (record_mode != 0) {
		    if (xwrite(1, tv.tv_CompBufAry[frame], block.b_Size) != block.b_Size)
			break;
		}
	    }
	    /*
	     * Display NFRAME-delayed frame
	     */
#if 1
	    ++tv.tv_FrameNo;
	    frame = tv.tv_FrameNo % NFRAME;
#endif
	    n = tv.tv_CompCountAry[frame];
	    if (n) {
		if (record_mode == 0) {
		    TVFrameDecompress(&tv, &xinfo, n);
		    FrameConvert(&tv, &xinfo);
		    if (head.h_Flags & THF_FILE)
			frameDelay(&tv);
		    FrameDisplay(&xinfo);
		}
		c = 'f';
	    } else {
		/*
		 * Frame lost at transmitter
		 */
		if (record_mode == 0) {
		    if (head.h_Flags & THF_FILE)
			frameDelay(&tv);
		}
		c = 'T';
	    }
	    if (VerboseOpt)
		write(2, &c, 1);
	    ++count;
#if 0
	    ++tv.tv_FrameNo;
#endif
	}
	if (VerboseOpt)
	    fprintf(stderr, "count %d frames\n", count);
	sleep(1);
    }
    return(0);
}

static
int 
xread(int fd, void *buf, int bytes)
{
    int t = 0;
    int r = 0;

    while (bytes && (r = read(fd, buf, bytes)) > 0) {
	bytes -= r;
	buf = (char *)buf + r;
	t += r;
    }
    if (t == 0)
	return(r);
    return(t);
}

static
int 
xwrite(int fd, void *buf, int bytes)
{
    int t = 0;
    int r = 0;

    while (bytes) {
	r = write(fd, buf, bytes);
	if (r >= 0) {
	    bytes -= r;
	    buf = (char *)buf + r;
	    t += r;
	    continue;
	}
	break;
    }
    if (t == 0)
	return(r);
    return(t);
}

static int
ConnectINetSocket(const char *host, int port)
{
    struct sockaddr_in sock;
    struct hostent *he;
    int ufd;

    /*
     * Construct internet domain socket
     */
    bzero(&sock, sizeof(sock));
    if ((ufd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
	return(-1);
    }
    sock.sin_family = AF_INET;
    sock.sin_port = htons(port);

    if (host == NULL) {
	sock.sin_addr.s_addr = 0;
    } else if (inet_aton(host, &sock.sin_addr) == 0) {
	if ((he = gethostbyname2(host, AF_INET)) == NULL) {
	    return(-1);
	}
	bcopy(he->h_addr, &sock.sin_addr, he->h_length);
    }

    /*
     * Bind to the interface/port
     */
    if (connect(ufd, (void *)&sock, sizeof(sock)) < 0) {
	close(ufd);
	ufd = -1;
    }
    return(ufd);
}

/*
 * The socket buffer has to be big so we can read() or write() an entire
 * frame without blocking.
 */
static void
setsockbuffer(int fd, int how)
{
    int bufsize = 512 * 1024;

    while (bufsize > 1024) {
	if (setsockopt(fd, SOL_SOCKET, how, &bufsize, sizeof(int)) == 0)
	    break;
	if (errno == ENOTSOCK || errno == EOPNOTSUPP)
	    break;
	bufsize -= 1024;
    }
    if (VerboseOpt)
	fprintf(stderr, "socket buffer: %d\n", bufsize);
}

static struct timeval Tv1;

static
void
frameDelayInit(void)
{
    gettimeofday(&Tv1, NULL);
}

static
void
frameDelay(TVCap *tv)
{
    struct timeval tv2;
    int us;

    gettimeofday(&tv2, NULL);
    Tv1.tv_usec += 1000000 / tv->tv_FPS;
    while (Tv1.tv_usec > 1000000) {
	Tv1.tv_usec -= 1000000;
	++Tv1.tv_sec;
    }
    us = (tv2.tv_usec + 1000000 - Tv1.tv_usec) +
	(tv2.tv_sec - Tv1.tv_sec - 1) * 1000000;
    if (us < 0)
	usleep(-us);
}

static void
setAlgorithm(TVHead *head, const char **chset)
{
    /*
     * Compression algorithm
     */
    switch(head->h_CAlgorithm) {
    case 1:
	TVFrameCompress = TVFrameCompress1;
	TVFrameDecompress = TVFrameDecompress1;
	break;
    case 2:
	TVFrameCompress = TVFrameCompress2;
	TVFrameDecompress = TVFrameDecompress2;
	break;
    case 3:
	TVFrameCompress = TVFrameCompress3;
	TVFrameDecompress = TVFrameDecompress3;
	break;
    case 4:
	TVFrameCompress = TVFrameCompress4;
	TVFrameDecompress = TVFrameDecompress4;
	break;
    default:
	fprintf(stderr, "Unknown compression algorithm, using #1\n");
	TVFrameCompress = TVFrameCompress1;
	TVFrameDecompress = TVFrameDecompress1;
	head->h_CAlgorithm = 1;
	break;
    }

    /*
     * Channel set.  *chset may be NULL, indicating that we should use the
     * default.  h_CHSet[0] may be 0 (empty string), indicating that we
     * should use the default.  h_CHSet[0] overrides *chset.  This works
     * as expected for all client and server modes.
     */
    head->h_CHSet[sizeof(head->h_CHSet)-1] = 0;
    if (head->h_CHSet[0] == 0 && *chset != NULL)
	snprintf(head->h_CHSet, sizeof(head->h_CHSet), "%s", *chset);
    if (head->h_CHSet[0] != 0)
	*chset = head->h_CHSet;
}

static void
blockFixup(TVBlock *b)
{
    b->b_Type = ntohl(b->b_Type);
    b->b_Info = ntohl(b->b_Info);
    b->b_Size = ntohl(b->b_Size);
}

static void
blockEncode(TVBlock *b)
{
    b->b_Type = htonl(b->b_Type);
    b->b_Info = htonl(b->b_Info);
    b->b_Size = htonl(b->b_Size);
}

#if 0
static
void
forkinputpipe(int fd)
{
    int fds[2];

    if (pipe(fds) < 0) {
	perror("pipe");
	exit(1);
    }
    if (fork() == 0) {
	close(fds[0]);
	if (fd != 0) {		/* fd becomes stdin of child */
	    dup2(fd, 0);
	    close(fd);
	}
	if (fds[1] != 1) {	/* stdout of child to pipe */
	    dup2(fds[1], 1);
	    close(fds[1]);
	}
	execl("/usr/bin/uncompress", "uncompress", NULL);
	_exit(1);
    }
    dup2(fds[0], fd);		/* set fd to become new stdin */
    close(fds[0]);
    close(fds[1]);
}

static
void
forkoutputpipe(int fd)
{
    int fds[2];

    if (pipe(fds) < 0) {
	perror("pipe");
	exit(1);
    }
    if (fork() == 0) {
	close(fds[1]);
	if (fd != 1) {		/* fd becomes stdout of child */
	    dup2(fd, 1);
	    close(fd);
	}
	if (fds[0] != 0) {	/* stdin of child to pipe */
	    dup2(fds[0], 0);
	    close(fds[0]);
	}
	execl("/usr/bin/compress", "compress", NULL);
	_exit(1);
    }
    dup2(fds[1], fd);		/* set fd to be new stdout */
    close(fds[0]);
    close(fds[1]);
}

static
void
forkoutputbuffer(int fd)
{
    int fds[2];

    if (pipe(fds) < 0) {
	perror("pipe");
	exit(1);
    }
    if (fork() == 0) {
	close(fds[1]);
	if (fd != 1) {		/* fd becomes stdout of child */
	    dup2(fd, 1);
	    close(fd);
	}
	if (fds[0] != 0) {	/* stdin of child to pipe */
	    dup2(fds[0], 0);
	    close(fds[0]);
	}
	execl("/bin/dd", "dd", "bs=1m", NULL);
	_exit(1);
    }
    dup2(fds[1], fd);		/* set fd to be new stdout */
    close(fds[0]);
    close(fds[1]);
}

#endif
