/*
 * TVCAP.C
 *
 * (c)Copyright 2002, Matthew Dillon, see COPYRIGHT file.
 * $Backplane: dtv/tvcap.c,v 1.14 2002/02/11 03:01:23 dillon Exp $
 */

#include "defs.h"

Prototype void TVCapInit(TVCap *tv, int shift, int channel, const char *chset);
Prototype void TVCapInitSlave(TVCap *tv, int xsize, int ysize, int shift, int channel, const char *chset);
Prototype int TVCapStart(TVCap *tv, char **emsg);
Prototype void TVCapStop(TVCap *tv);
Prototype void TVCapSetChannel(TVCap *tv);
Prototype int TVFrameGet(TVCap *tv);

static void framesig(int sigNo);
static CHSet *findChannelSet(const char *name);

static int FrameCount;

void
TVCapInitSlave(TVCap *tv, int xsize, int ysize, int shift, int channel, const char *chset)
{
    int i;
    CHSet *cs = findChannelSet(chset);

    if (cs == NULL)
	fatal("Channel set '%s' could not be found", chset);

    bzero(tv, sizeof(TVCap));
    tv->tv_XSize = (xsize < 0) ? cs->cs_XSize : xsize;
    tv->tv_YSize = (ysize < 0) ? cs->cs_YSize : ysize;
    tv->tv_Shift = (shift < 0) ? 0 : shift;	/* capture can't go negative */
    tv->tv_CapFd = -1;
    tv->tv_TuneFd = -1;
    tv->tv_CapBuf = MAP_FAILED;
    tv->tv_CapBufSize = tv->tv_XSize * tv->tv_YSize * MAXPIXSIZE;
    tv->tv_CompBufSize = tv->tv_CapBufSize * 2;
    tv->tv_ChanSet = cs->cs_ChanSet;
    tv->tv_AudioIn = METEOR_INPUT_DEV1;	/* tuner */
    tv->tv_TunerIn = AUDIO_TUNER;
    tv->tv_VideoIn = -1;
    tv->tv_VideoFmt = cs->cs_IForm;
    tv->tv_FPS = cs->cs_FrameRate;
    tv->tv_Channel = channel;
    tv->tv_Source = SOURCE_CMPBUF;
    tv->tv_CmpBuf = malloc(tv->tv_CapBufSize);	/* source 2 */
    for (i = 0; i < NFRAME; ++i)
	tv->tv_CompBufAry[i] = malloc(tv->tv_CompBufSize);
#if 1
    tv->tv_DevOutputFormat = METEOR_GEO_RGB24;
    tv->tv_PixSize = 4;
#endif
#if 0
    tv->tv_DevOutputFormat = METEOR_GEO_RGB16;
    tv->tv_PixSize = 2;
#endif
    bzero(tv->tv_CmpBuf, tv->tv_CapBufSize);
    signal(SIGIO, framesig);
}

void
TVCapInit(TVCap *tv, int shift, int channel, const char *chset)
{
    TVCapInitSlave(tv, -1, -1, shift, channel, chset);
}

int
TVCapStart(TVCap *tv, char **emsg)
{
    int error = -1;
    bktr_cmd_t v;

    if ((tv->tv_CapFd = open("/dev/bktr0", O_RDONLY)) < 0) {
	asprintf(emsg, "Unable to open /dev/bktr: %s", strerror(errno));
	goto done;
    }
    fcntl(tv->tv_CapFd, F_SETFD, 1);
    if ((tv->tv_TuneFd = open("/dev/tuner0", O_RDONLY)) < 0) {
	asprintf(emsg, "Unable to open /dev/tuner: %s", strerror(errno));
	goto done;
    }
    fcntl(tv->tv_TuneFd, F_SETFD, 1);
    if (ioctl(tv->tv_CapFd, BT848SFMT, &tv->tv_VideoFmt) < 0) {
	asprintf(emsg, "ioctl(BT848SFMT) failed: %s", strerror(errno));
	goto done;
    }
#if 0
    bzero(&cs, sizeof(cs));
    cs.index = 1;
    while (ioctl(tv->tv_TuneFd, TVTUNER_GETCHNLSET, &cs) == 0) {
	if (strcmp(cs.name, "cableirc") == 0) {
	    tv->tv_ChanSet = cs.index;
	    ...
	}
	++cs.index;
    }
#endif
    if (ioctl(tv->tv_TuneFd, TVTUNER_SETTYPE, &tv->tv_ChanSet) < 0) {
	asprintf(emsg, "ioctl(TVTUNER_SETTYPE) failed: %s", strerror(errno));
	goto done;
    }
    if (ioctl(tv->tv_CapFd, METEORSINPUT, &tv->tv_AudioIn) < 0) {
	asprintf(emsg, "ioctl(METEORSINPUT) failed: %s", strerror(errno));
	goto done;
    }
    if (ioctl(tv->tv_TuneFd, BT848_SAUDIO, &tv->tv_TunerIn) < 0) {
	asprintf(emsg, "ioctl(BT848_SAUDIO) failed: %s", strerror(errno));
	goto done;
    }
    {
	struct meteor_geomet mg;
	bzero(&mg, sizeof(mg));
	mg.columns = tv->tv_XSize;
	mg.rows    = tv->tv_YSize;
	mg.frames  = 1;
	mg.oformat = tv->tv_DevOutputFormat;
	if (ioctl(tv->tv_CapFd, METEORSETGEO, &mg) < 0) {
	    asprintf(emsg, "ioctl(METEORSETGEO) failed: %s", strerror(errno));
	    goto done;
	}
    }

    v = SIGIO;
    if (ioctl(tv->tv_CapFd, METEORSSIGNAL, &v) < 0) {
	asprintf(emsg, "ioctl(METEORSSIGNAL) failed: %s", strerror(errno));
	goto done;
    }
    tv->tv_CapBuf = mmap(NULL, tv->tv_CapBufSize, PROT_READ, 
			MAP_SHARED, tv->tv_CapFd, 0);
    tv->tv_Source = SOURCE_CAPBUF;
    if (tv->tv_CapBuf == MAP_FAILED) {
	asprintf(emsg, "Unable to mmap() frame buffer: %s", strerror(errno));
	goto done;
    }
    TVCapSetChannel(tv);
    v = METEOR_CAP_CONTINOUS;
    if (ioctl(tv->tv_CapFd, METEORCAPTUR, &v) < 0) {
	asprintf(emsg, "ioctl(METEORCAPTUR) failed: %s", strerror(errno));
	goto done;
    }
    error = 0;
done:
    if (error)
	TVCapStop(tv);
    return(error);
}

void
TVCapStop(TVCap *tv)
{
    if (tv->tv_CapBuf != MAP_FAILED) {
	munmap(tv->tv_CapBuf, tv->tv_CapBufSize);
	tv->tv_CapBuf = MAP_FAILED;
    }
    if (tv->tv_CmpBuf != NULL) {
	free(tv->tv_CmpBuf);
	tv->tv_CmpBuf = NULL;
    }
    if (tv->tv_CapFd >= 0) {
	close(tv->tv_CapFd);
	tv->tv_CapFd = -1;
    }
    if (tv->tv_TuneFd >= 0) {
	close(tv->tv_TuneFd);
	tv->tv_TuneFd = -1;
    }
    signal(SIGIO, SIG_IGN);
}

void
TVCapSetChannel(TVCap *tv)
{
    if (ioctl(tv->tv_TuneFd, TVTUNER_SETCHNL, &tv->tv_Channel) < 0)
	fprintf(stderr, "Unable to set channel\n");
    /* reset audio mute */
}

int
TVFrameGet(TVCap *tv)
{
    int n = tv->tv_XSize * tv->tv_YSize;
    int mask;

    mask = sigblock(SIGIO);
    while (FrameCount == 0)
	sigpause(0);
    --FrameCount;
    sigsetmask(mask);
    if (FrameCount == 0)
	return(n * 4);	/* good frame */
    else
	return(0);	/* missed a frame */
}

static void
framesig(int sigNo)
{
    ++FrameCount;
}

/*
 * XXX This needs a lot of work.  I obviously do not have the 
 * correct dimensions or frame rates for all the configurations.
 *
 * BT848_IFORM_F_NTSCM		640x480/30
 * BT848_IFORM_F_NTSCJ		640x480/30
 * BT848_IFORM_F_PALM		640x480/30
 *
 * BT848_IFORM_F_PALBDGHI	768x576/25
 * BT848_IFORM_F_PALN		768x576/25
 * BT848_IFORM_F_SECAM		768x576/25
 *
 * BT848_IFORM_F_RSVD
 */

static CHSet CHSetAry[] = {
    { "cableirc", CHNLSET_CABLEIRC, BT848_IFORM_F_NTSCM, 640, 480, 30 },
    { "nabcst",	  CHNLSET_NABCST,   BT848_IFORM_F_NTSCM, 640, 480, 30 },
    { "cablehrc", CHNLSET_CABLEHRC, BT848_IFORM_F_NTSCM, 640, 480, 30 },
    { "weurope",  CHNLSET_WEUROPE,  BT848_IFORM_F_NTSCM, 640, 480, 30 },
    { "jpnbcst",  CHNLSET_JPNBCST,  BT848_IFORM_F_NTSCM, 640, 480, 30 },
    { "jpncable", CHNLSET_JPNCABLE, BT848_IFORM_F_NTSCM, 640, 480, 30 },
    { "xussr",	  CHNLSET_XUSSR,    BT848_IFORM_F_NTSCM, 640, 480, 30 },
    { "australia",CHNLSET_AUSTRALIA,BT848_IFORM_F_NTSCM, 640, 480, 30 },
    { "france",	  CHNLSET_FRANCE,   BT848_IFORM_F_NTSCM, 640, 480, 30 }
};

static CHSet *
findChannelSet(const char *name)
{
    CHSet *cs = &CHSetAry[0];

    while (name && cs < &CHSetAry[arysize(CHSetAry)]) {
	if (strcasecmp(name, cs->cs_Name) == 0)
	    break;
	++cs;
    }
    if (cs == &CHSetAry[arysize(CHSetAry)])
	cs = NULL;
    return(cs);
}

